from diofant import cos, pretty, sin
from diofant.abc import x
from diofant.printing.pretty.pretty import PrettyPrinter


__all__ = ()


def test_render():
    ustr = \
        """\
     3     5      \n\
    x     x     ⎛ \n\
x - ── + ─── + O⎝x\n\
    6    120      \n\
\n\
  \n\
6⎞\n\
 ⎠\n\
  \
"""
    assert pretty(sin(x).series(x), num_columns=20,
                  use_unicode=True) == ustr


def test_stringpict():
    p = PrettyPrinter()

    p1 = p._print(sin(x))
    p2 = p._print(sin(x))
    assert p1 == p2
    assert hash(p1) == hash(p2)

    p3 = p._print(cos(x))
    assert p1 != p3
    assert p1 != 0

    assert len(p1) == 1


def test_dumb_term(capsys):
    print(pretty(sin(x)))
    assert capsys.readouterr().out == 'sin(x)\n'


def test_ncol():
    res = pretty(sin(x).series(x, n=20), num_columns=0)
    ans = """\
     3     5     7       9        11          13             15               \n\
    x     x     x       x        x           x              x                x\n\
x - ── + ─── - ──── + ────── - ──────── + ────────── - ───────────── + ───────\n\
    6    120   5040   362880   39916800   6227020800   1307674368000   3556874\n\
\n\
17                 19                 \n\
                  x              ⎛ 20⎞\n\
──────── - ────────────────── + O⎝x  ⎠\n\
28096000   121645100408832000         \
"""
    assert res == ans



"""Computational algebraic field theory."""

import functools
import math

import mpmath

from ..config import query
from ..core import (Add, Dummy, E, GoldenRatio, I, Integer, Mul, Rational,
                    cacheit, pi)
from ..core.exprtools import Factors
from ..core.function import _mexpand, count_ops
from ..core.sympify import sympify
from ..domains import QQ, AlgebraicField
from ..functions import (Abs, conjugate, cos, exp_polar, im, re, root, sin,
                         sqrt, tan)
from ..ntheory import divisors, factorint
from ..simplify.radsimp import _split_gcd
from ..simplify.simplify import _is_sum_surds
from ..utilities import lambdify, numbered_symbols, sift
from ..utilities.iterables import uniq
from .orthopolys import chebyshevt_poly
from .polyerrors import NotAlgebraic
from .polytools import (Poly, PurePoly, degree, factor_list, groebner, lcm,
                        parallel_poly_from_expr, resultant)
from .rootoftools import RootOf
from .specialpolys import cyclotomic_poly


__all__ = 'minimal_polynomial', 'primitive_element', 'field_isomorphism'


def _choose_factor(factors, x, v, dom=QQ, prec=200, bound=5):
    """
    Return a factor having root ``v``
    It is assumed that one of the factors has root ``v``.

    """
    if isinstance(factors[0], tuple):
        factors = [f[0] for f in factors]
    if len(factors) == 1:
        return factors[0]

    points = {x: v}
    symbols = dom.symbols if hasattr(dom, 'symbols') else []
    t = QQ(1, 10)

    for n in range(bound**len(symbols)):
        prec1 = 10
        n_temp = n
        for s in symbols:
            points[s] = n_temp % bound
            n_temp = n_temp // bound

        while True:
            candidates = []
            eps = t**(prec1 // 2)
            for f in factors:
                if abs(f.as_expr().evalf(prec1, points, strict=False)) < eps:
                    candidates.append(f)
            if candidates:
                factors = candidates
            if len(factors) == 1:
                return factors[0]
            if prec1 > prec:
                break
            prec1 *= 2

    raise NotImplementedError(f'multiple candidates for the minimal polynomial of {v}')


def _separate_sq(p):
    """
    Helper function for ``_minimal_polynomial_sq``.

    It selects a rational ``g`` such that the polynomial ``p``
    consists of a sum of terms whose surds squared have gcd equal to ``g``
    and a sum of terms with surds squared prime with ``g``;
    then it takes the field norm to eliminate ``sqrt(g)``

    See simplify.simplify.split_surds and polytools.sqf_norm.

    Examples
    ========

    >>> p = -x + sqrt(2) + sqrt(3) + sqrt(7)
    >>> p = _separate_sq(p)
    >>> p
    -x**2 + 2*sqrt(3)*x + 2*sqrt(7)*x - 2*sqrt(21) - 8
    >>> p = _separate_sq(p)
    >>> p
    -x**4 + 4*sqrt(7)*x**3 - 32*x**2 + 8*sqrt(7)*x + 20
    >>> p = _separate_sq(p)
    >>> p
    -x**8 + 48*x**6 - 536*x**4 + 1728*x**2 - 400

    """
    def is_sqrt(expr):
        return expr.is_Pow and expr.exp == Rational(1, 2)

    p = p.doit()

    # p = c1*sqrt(q1) + ... + cn*sqrt(qn) -> a = [(c1, q1), .., (cn, qn)]
    a = []
    for y in p.args:
        if not y.is_Mul:
            if is_sqrt(y):
                a.append((Integer(1), y**2))
            elif y.is_Atom:
                a.append((y, Integer(1)))
            else:
                raise NotImplementedError
        else:
            sifted = sift(y.args, is_sqrt)
            a.append((Mul(*sifted[False]), Mul(*sifted[True])**2))
    a.sort(key=lambda z: z[1])
    if a[-1][1] == 1:
        # there are no surds
        return p
    surds = [z for y, z in a]
    for i, si in enumerate(surds):  # pragma: no branch
        if si != 1:
            break
    _, b1, _ = _split_gcd(*surds[i:])
    a1 = []
    a2 = []
    for y, z in a:
        if z in b1:
            a1.append(y*sqrt(z))
        else:
            a2.append(y*sqrt(z))
    p1 = Add(*a1)
    p2 = Add(*a2)
    return _mexpand(p1**2) - _mexpand(p2**2)


def _minimal_polynomial_sq(p, n, x):
    """
    Returns the minimal polynomial for the ``nth-root`` of a sum of surds
    or ``None`` if it fails.

    Parameters
    ==========

    p : sum of surds
    n : positive integer
    x : variable of the returned polynomial

    Examples
    ========

    >>> q = 1 + sqrt(2) + sqrt(3)
    >>> _minimal_polynomial_sq(q, 3, x)
    x**12 - 4*x**9 - 4*x**6 + 16*x**3 - 8

    """
    p = sympify(p)
    n = sympify(n)
    assert n.is_Integer
    assert n > 1
    assert _is_sum_surds(p)
    pn = root(p, n)
    # eliminate the square roots
    p -= x
    while 1:
        p1 = _separate_sq(p)
        if p1 is p:
            p = p1.subs({x: x**n})
            break
        p = p1

    # by construction `p` has root `pn`
    # the minimal polynomial is the factor vanishing in x = pn
    factors = factor_list(p)[1]

    return _choose_factor(factors, x, pn)


def _minpoly_op_algebraic_element(op, ex1, ex2, x, dom, mp1=None, mp2=None):
    """
    Return the minimal polynomial for ``op(ex1, ex2)``.

    Parameters
    ==========

    op : operation ``Add`` or ``Mul``
    ex1, ex2 : expressions for the algebraic elements
    x : indeterminate of the polynomials
    dom: ground domain
    mp1, mp2 : minimal polynomials for ``ex1`` and ``ex2`` or None

    Examples
    ========

    >>> p1 = sqrt(sqrt(2) + 1)
    >>> p2 = sqrt(sqrt(2) - 1)
    >>> _minpoly_op_algebraic_element(Mul, p1, p2, x, QQ)
    x - 1
    >>> q1 = sqrt(y)
    >>> q2 = 1 / y
    >>> _minpoly_op_algebraic_element(Add, q1, q2, x, QQ.inject(y).field)
    x**2*y**2 - 2*x*y - y**3 + 1

    References
    ==========

    * https://en.wikipedia.org/wiki/Resultant
    * I.M. Isaacs, Proc. Amer. Math. Soc. 25 (1970), 638
      "Degrees of sums in a separable field extension".

    """
    y = Dummy(str(x))
    if mp1 is None:
        mp1 = _minpoly_compose(ex1, x, dom)
    if mp2 is None:
        mp2 = _minpoly_compose(ex2, y, dom)
    else:
        mp2 = mp2.subs({x: y})

    if op is Add:
        # mp1a = mp1.subs({x: x - y})
        (p1, p2), _ = parallel_poly_from_expr((mp1, x - y), x, y)
        r = p1.compose(p2)
        mp1a = r.as_expr()

    elif op is Mul:
        mp1a = _muly(mp1, x, y)
    else:
        raise NotImplementedError('option not available')

    r = resultant(mp1a, mp2, gens=[y, x])

    deg1 = degree(mp1, x)
    deg2 = degree(mp2, y)
    if op is Mul and deg1 == 1 or deg2 == 1:
        # if deg1 = 1, then mp1 = x - a; mp1a = x - y - a;
        # r = mp2(x - a), so that `r` is irreducible
        return r

    r = r.as_poly(x, domain=dom)
    _, factors = r.factor_list()
    res = _choose_factor(factors, x, op(ex1, ex2), dom)
    return res.as_expr()


def _invertx(p, x):
    """Returns ``expand_mul(x**degree(p, x)*p.subs({x: 1/x}))``."""
    (p1,) = parallel_poly_from_expr((p,), x)[0]

    n = degree(p1)
    a = [c * x**(n - i) for (i,), c in p1.terms()]
    return Add(*a)


def _muly(p, x, y):
    """Returns ``_mexpand(y**deg*p.subs({x:x / y}))``."""
    (p1,) = parallel_poly_from_expr((p,), x)[0]

    n = degree(p1)
    a = [c * x**i * y**(n - i) for (i,), c in p1.terms()]
    return Add(*a)


def _minpoly_pow(ex, pw, x, dom):
    """
    Returns ``minimal_polynomial(ex**pw)``

    Parameters
    ==========

    ex : algebraic element
    pw : rational number
    x : indeterminate of the polynomial
    dom: ground domain

    Examples
    ========

    >>> p = sqrt(1 + sqrt(2))
    >>> _minpoly_pow(p, 2, x, QQ)
    x**2 - 2*x - 1
    >>> minimal_polynomial(p**2)(x)
    x**2 - 2*x - 1
    >>> _minpoly_pow(y, Rational(1, 3), x, QQ.inject(y).field)
    x**3 - y
    >>> minimal_polynomial(cbrt(y))(x)
    x**3 - y

    """
    pw = sympify(pw)
    mp = _minpoly_compose(ex, x, dom)
    if not pw.is_rational:
        raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")
    if pw < 0:
        if mp == x:
            raise ZeroDivisionError(f'{ex} is zero')
        mp = _invertx(mp, x)
        if pw == -1:
            return mp
        pw = -pw
        ex = 1/ex

    y = Dummy(str(x))
    mp = mp.subs({x: y})
    n, d = pw.as_numer_denom()
    res = resultant(mp, x**d - y**n, gens=[y]).as_poly(x, domain=dom)
    _, factors = res.factor_list()
    res = _choose_factor(factors, x, ex**pw, dom)
    return res.as_expr()


def _minpoly_add(x, dom, *a):
    """Returns ``minimal_polynomial(Add(*a), dom)``."""
    mp = _minpoly_op_algebraic_element(Add, a[0], a[1], x, dom)
    p = a[0] + a[1]
    for px in a[2:]:
        mp = _minpoly_op_algebraic_element(Add, p, px, x, dom, mp1=mp)
        p = p + px
    return mp


def _minpoly_mul(x, dom, *a):
    """Returns ``minimal_polynomial(Mul(*a), dom)``."""
    mp = _minpoly_op_algebraic_element(Mul, a[0], a[1], x, dom)
    p = a[0] * a[1]
    for px in a[2:]:
        mp = _minpoly_op_algebraic_element(Mul, p, px, x, dom, mp1=mp)
        p = p * px
    return mp


def _minpoly_sin(ex, x):
    """
    Returns the minimal polynomial of ``sin(ex)``
    see https://mathworld.wolfram.com/TrigonometryAngles.html

    """
    c, a = ex.args[0].as_coeff_Mul()
    if a is pi:
        n = c.denominator
        q = sympify(n)
        if q.is_prime:
            # for a = pi*p/q with q odd prime, using chebyshevt
            # write sin(q*a) = mp(sin(a))*sin(a);
            # the roots of mp(x) are sin(pi*p/q) for p = 1,..., q - 1
            a = chebyshevt_poly(n, polys=True).all_coeffs()
            return Add(*[x**(n - i - 1)*a[n - i] for i in range(n)])
        if c.numerator == 1:
            if q == 9:
                return 64*x**6 - 96*x**4 + 36*x**2 - 3

        if n % 2 == 1:
            # for a = pi*p/q with q odd, use
            # sin(q*a) = 0 to see that the minimal polynomial must be
            # a factor of chebyshevt_poly(n)
            a = chebyshevt_poly(n, polys=True).all_coeffs()
            a = [x**(n - i)*a[n - i] for i in range(n + 1)]
            r = Add(*a)
            _, factors = factor_list(r)
            res = _choose_factor(factors, x, ex)
            return res

        expr = sqrt((1 - cos(2*c*pi))/2)
        return _minpoly_compose(expr, x, QQ)

    raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")


def _minpoly_cos(ex, x):
    """
    Returns the minimal polynomial of ``cos(ex)``
    see https://mathworld.wolfram.com/TrigonometryAngles.html

    """
    c, a = ex.args[0].as_coeff_Mul()
    if a is pi:
        if c.numerator == 1:
            if c.denominator == 7:
                return 8*x**3 - 4*x**2 - 4*x + 1
            elif c.denominator == 9:
                return 8*x**3 - 6*x - 1
        elif c.numerator == 2:
            q = sympify(c.denominator)
            if q.is_prime:
                s = _minpoly_sin(ex, x)
                return _mexpand(s.subs({x: sqrt((1 - x)/2)}))

        # for a = pi*p/q, cos(q*a) =T_q(cos(a)) = (-1)**p
        n = int(c.denominator)
        a = chebyshevt_poly(n, polys=True).all_coeffs()
        a = [x**(n - i)*a[n - i] for i in range(n + 1)]
        r = Add(*a) - (-1)**c.numerator
        _, factors = factor_list(r)
        return _choose_factor(factors, x, ex)

    raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")


def _minpoly_tan(ex, x):
    """Returns the minimal polynomial of ``tan(ex)``."""
    c, a = ex.args[0].as_coeff_Mul()
    if a is pi and c.is_Rational:
        c *= 2
        n = c.denominator
        a = n if c.numerator % 2 == 0 else 1
        terms = []
        for k in range((c.numerator + 1) % 2, n + 1, 2):
            terms.append(a*x**k)
            a = -(a*(n - k - 1)*(n - k)) // ((k + 1)*(k + 2))

        r = Add(*terms)
        _, factors = factor_list(r)
        return _choose_factor(factors, x, ex)

    raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")


def _minpoly_exp(ex, x):
    """Returns the minimal polynomial of ``exp(ex)``."""
    c, a = ex.exp.as_coeff_Mul()
    q = sympify(c.denominator)
    if a == I*pi:
        if c.numerator in (1, -1):
            if q == 3:
                return x**2 - x + 1
            if q == 4:
                return x**4 + 1
            if q == 6:
                return x**4 - x**2 + 1
            if q == 8:
                return x**8 + 1
            if q == 9:
                return x**6 - x**3 + 1
            if q == 10:
                return x**8 - x**6 + x**4 - x**2 + 1
            if q.is_prime:
                s = 0
                for i in range(q):
                    s += (-x)**i
                return s

        # x**(2*q) = product(factors)
        factors = [cyclotomic_poly(i, x) for i in divisors(2*q)]
        return _choose_factor(factors, x, ex)
    raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")


def _minpoly_rootof(ex, x):
    """Returns the minimal polynomial of a ``RootOf`` object."""
    domain = ex.poly.domain
    if domain.is_IntegerRing:
        return ex.poly(x)
    else:
        return ex.poly.sqf_norm()[-1](x)


def _minpoly_compose(ex, x, dom):
    """
    Computes the minimal polynomial of an algebraic element
    using operations on minimal polynomials

    Examples
    ========

    >>> minimal_polynomial(sqrt(2) + 3*Rational(1, 3), method='compose')(x)
    x**2 - 2*x - 1
    >>> minimal_polynomial(sqrt(y) + 1/y, method='compose')(x)
    x**2*y**2 - 2*x*y - y**3 + 1

    """
    if ex.is_Rational:
        return ex.denominator*x - ex.numerator
    if ex is I:
        return x**2 + 1
    if ex is GoldenRatio:
        return x**2 - x - 1
    if ex == exp_polar(0):
        return x - 1
    if hasattr(dom, 'symbols') and ex in dom.symbols:
        return x - ex

    if dom.is_RationalField and _is_sum_surds(ex):
        # eliminate the square roots
        ex -= x
        while 1:
            ex1 = _separate_sq(ex)
            if ex1 is ex:
                return ex
            else:
                ex = ex1

    if ex.is_Add:
        res = _minpoly_add(x, dom, *sorted(ex.args, key=count_ops, reverse=True))
    elif ex.is_Mul:
        f = Factors(ex).factors
        r = sift(f.items(), lambda itx: itx[0].is_Rational and itx[1].is_Rational)
        if r[True] and dom == QQ:
            ex1 = Mul(*[bx**ex for bx, ex in r[False] + r[None]])
            r1 = r[True]
            dens = [y.denominator for _, y in r1]
            lcmdens = functools.reduce(lcm, dens, 1)
            nums = [base**(y.numerator*lcmdens // y.denominator) for base, y in r1]
            ex2 = Mul(*nums)
            mp1 = minimal_polynomial(ex1)(x)
            # use the fact that in Diofant canonicalization products of integers
            # raised to rational powers are organized in relatively prime
            # bases, and that in ``base**(n/d)`` a perfect power is
            # simplified with the root
            mp2 = ex2.denominator*x**lcmdens - ex2.numerator
            ex2 = Mul(*[bx**ex for bx, ex in r1])
            res = _minpoly_op_algebraic_element(Mul, ex1, ex2, x, dom, mp1=mp1, mp2=mp2)
        else:
            res = _minpoly_mul(x, dom, *sorted(ex.args, key=count_ops, reverse=True))
    elif ex.is_Pow:
        if ex.base is E:
            res = _minpoly_exp(ex, x)
        else:
            res = _minpoly_pow(ex.base, ex.exp, x, dom)
    elif isinstance(ex, sin):
        res = _minpoly_sin(ex, x)
    elif isinstance(ex, cos):
        res = _minpoly_cos(ex, x)
    elif isinstance(ex, tan):
        res = _minpoly_tan(ex, x)
    elif isinstance(ex, RootOf) and ex.poly.domain.is_Numerical:
        res = _minpoly_rootof(ex, x)
    elif isinstance(ex, conjugate):
        res = _minpoly_compose(ex.args[0], x, dom)
    elif isinstance(ex, Abs):
        res = _minpoly_compose(sqrt(ex.args[0]*ex.args[0].conjugate()), x, dom)
    elif isinstance(ex, re):
        res = _minpoly_compose((ex.args[0] + ex.args[0].conjugate())/2, x, dom)
    elif isinstance(ex, im):
        res = _minpoly_compose((ex.args[0] - ex.args[0].conjugate())/2/I, x, dom)
    else:
        raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic element")
    return res


@cacheit
def minimal_polynomial(ex, method=None, **args):
    """
    Computes the minimal polynomial of an algebraic element.

    Parameters
    ==========

    ex : algebraic element expression
    method : str, optional
        If ``compose``, the minimal polynomial of the subexpressions
        of ``ex`` are computed, then the arithmetic operations on them are
        performed using the resultant and factorization.  If ``groebner``,
        a bottom-up algorithm, using Gröbner bases is used.
        Defaults are determined by :func:`~diofant.config.setup`.
    domain : Domain, optional
        If no ground domain is given, it will be generated automatically
        from the expression.

    Examples
    ========

    >>> minimal_polynomial(sqrt(2))(x)
    x**2 - 2
    >>> minimal_polynomial(sqrt(2), domain=QQ.algebraic_field(sqrt(2)))(x)
    x - sqrt(2)
    >>> minimal_polynomial(sqrt(2) + sqrt(3))(x)
    x**4 - 10*x**2 + 1
    >>> minimal_polynomial(solve(x**3 + x + 3)[0][x])(x)
    x**3 + x + 3
    >>> minimal_polynomial(sqrt(y))(x)
    x**2 - y

    """
    if method is None:
        method = query('minpoly_method')
    _minpoly_methods = {'compose': _minpoly_compose, 'groebner': minpoly_groebner}
    try:
        _minpoly = _minpoly_methods[method]
    except KeyError as exc:
        raise ValueError(f"'{method}' is not a valid algorithm "
                         'for computing minimal polynomial') from exc

    ex = sympify(ex)
    if ex.is_number:
        # not sure if it's always needed but try it for numbers (issue sympy/sympy#8354)
        ex = _mexpand(ex, recursive=True)

    x = Dummy('x')
    domain = args.get('domain',
                      QQ.inject(*ex.free_symbols).field if ex.free_symbols else QQ)

    result = _minpoly(ex, x, domain)
    _, factors = factor_list(result, x, domain=domain)
    result = _choose_factor(factors, x, ex, dom=domain)
    result = result.primitive()[1]

    return PurePoly(result, x, domain=domain)


def minpoly_groebner(ex, x, domain):
    """
    Computes the minimal polynomial of an algebraic number
    using Gröbner bases

    Examples
    ========

    >>> minimal_polynomial(sqrt(2) + 1, method='groebner')(x)
    x**2 - 2*x - 1

    References
    ==========

    * :cite:`Adams1994intro`

    """
    generator = numbered_symbols('a', cls=Dummy)
    mapping, symbols = {}, {}

    def update_mapping(ex, exp, base=None):
        if ex in mapping:
            return symbols[ex]

        a = next(generator)
        symbols[ex] = a

        if base is not None:
            mapping[ex] = a**exp + base
        else:
            mapping[ex] = exp.as_expr(a)

        return a

    def bottom_up_scan(ex):
        if ex.is_Atom:
            if ex is I:
                return update_mapping(ex, 2, 1)
            elif ex is GoldenRatio:
                return bottom_up_scan(ex.expand(func=True))
            elif ex.is_Rational:
                return ex
            elif ex.is_Symbol:
                return ex
        elif ex.is_Add or ex.is_Mul:
            return ex.func(*[bottom_up_scan(g) for g in ex.args])
        elif ex.is_Pow:
            if ex.exp.is_Rational:
                base, exp = ex.base, ex.exp
                if exp.is_nonnegative:
                    if exp.is_noninteger:
                        base, exp = base**exp.numerator, Rational(1, exp.denominator)
                    base = bottom_up_scan(base)
                else:
                    bmp = PurePoly(minpoly_groebner(1/base, x, domain=domain), x)
                    base, exp = update_mapping(1/base, bmp), -exp
                return update_mapping(ex, exp.denominator, -base**exp.numerator)
        elif isinstance(ex, RootOf) and ex.poly.domain.is_Numerical:
            if ex.poly.domain.is_IntegerRing:
                return update_mapping(ex, ex.poly)
            else:
                return update_mapping(ex, ex.poly.sqf_norm()[-1])
        elif isinstance(ex, conjugate):
            return update_mapping(ex, minimal_polynomial(ex.args[0], domain=domain,
                                                         method='groebner'))
        elif isinstance(ex, Abs):
            return bottom_up_scan(sqrt(ex.args[0]*ex.args[0].conjugate()))
        elif isinstance(ex, re):
            return bottom_up_scan((ex.args[0] + ex.args[0].conjugate())/2)
        elif isinstance(ex, im):
            return bottom_up_scan((ex.args[0] - ex.args[0].conjugate())/2/I)

        raise NotAlgebraic(f"{ex} doesn't seem to be an algebraic number")

    if ex.is_Pow and ex.exp.is_negative:
        n, d = Integer(1), bottom_up_scan(1/ex)
    else:
        n, d = bottom_up_scan(ex), Integer(1)

    F = [d*x - n] + list(mapping.values())
    G = groebner(F, *(list(symbols.values()) + [x]), order='lex', domain=domain)

    return G[-1]  # by construction G[-1] has root `ex`


def primitive_element(extension, **args):
    """Construct a common number field for all extensions.

    References
    ==========

    * :cite:`Yokoyama1989primitive`
    * :cite:`Arno1996alg`

    """
    if not extension:
        raise ValueError("can't compute primitive element for empty extension")

    extension = list(uniq(extension))

    x = Dummy('x')
    domain = args.get('domain', QQ)
    F = [minimal_polynomial(e, domain=domain) for e in extension]
    Y = [p.gen for p in F]

    for u in range(1, (len(F) - 1)*math.prod(f.degree() for f in F) + 1):
        coeffs = [u**n for n in range(len(Y))]
        f = x - sum(c*y for c, y in zip(coeffs, Y))

        *H, g = groebner(F + [f], *(Y + [x]), domain=domain)

        for i, (h, y) in enumerate(zip(H, Y)):
            H[i] = (y - h).eject(*Y).retract(field=True)
            if not (H[i].domain.is_RationalField or H[i].domain.is_AlgebraicField):
                break  # G is not a triangular set
            H[i] = H[i].set_domain(domain)
        else:
            g = g.eject(*Y).set_domain(domain)
            break
    else:
        if len(F) == 1:
            g, coeffs, H = F[0].replace(x), [Integer(1)], [x.as_poly(domain=domain)]
        else:  # pragma: no cover
            raise RuntimeError('run out of coefficient configurations')

    _, factors = factor_list(g, domain=domain)
    t = sum(c*e for c, e in zip(coeffs, extension))
    g = _choose_factor(factors, x, t, dom=domain)

    H = [h.rem(g).rep.all_coeffs() for h in H]

    _, g = PurePoly(g).clear_denoms(convert=True)

    if g.LC() != 1:
        for d in divisors(g.LC())[1:]:  # pragma: no branch
            new_g = g.compose((g.gen/d).as_poly())*d**g.degree()//d
            _, new_g = new_g.monic().clear_denoms(convert=True)
            if new_g.LC() == 1:
                g = new_g
                H = [[c/d**n for n, c in enumerate(h)] for h in H]
                coeffs = [c*d for c in coeffs]
                break

    return g, list(coeffs), H


def field_isomorphism_pslq(a, b):
    """Construct field isomorphism using PSLQ algorithm."""
    if not all(_.domain.is_RationalField and _.ext.is_real for _ in (a, b)):
        raise NotImplementedError("PSLQ doesn't support complex coefficients")

    f = a.minpoly
    x = f.gen

    g = b.minpoly.replace(x)
    m = g.degree()

    a, b = a.ext, b.ext

    for n in mpmath.libmp.libintmath.giant_steps(32, 256):  # pragma: no branch
        with mpmath.workdps(n):
            A, B = lambdify((), [a, b], 'mpmath')()
            basis = [B**i for i in range(m)] + [A]
            coeffs = mpmath.pslq(basis, maxcoeff=10**10, maxsteps=10**3)

        if coeffs:
            assert coeffs[-1]  # basis[:-1] elements are linearly independent

            h = -Poly(coeffs[:-1], x, field=True).quo_ground(coeffs[-1])

            if f.compose(h).rem(g).is_zero:
                return h.rep.all_coeffs()
        else:
            break


def field_isomorphism_factor(a, b):
    """Construct field isomorphism via factorization."""
    p = a.minpoly.set_domain(b)
    _, factors = p.factor_list()

    for f, _ in factors:
        if f.degree() == 1:
            root = -f.rep[(0,)]/f.rep[(1,)]

            if (a.ext - b.to_expr(root)).evalf(chop=True) == 0:
                return root.rep.all_coeffs()


def field_isomorphism(a, b, **args):
    """Construct an isomorphism between two number fields."""
    if not all(isinstance(_, AlgebraicField) for _ in (a, b)):
        raise ValueError(f'Arguments should be algebraic fields, got {a} and {b}')

    if a == b:
        return a.unit.rep.all_coeffs()

    n = a.minpoly.degree()
    m = b.minpoly.degree()

    if a.domain == b.domain:
        if m % n:
            return
        elif a.domain.is_RationalField:
            da = a.minpoly.discriminant()
            db = b.minpoly.discriminant()
            k = m // n

            for p, q in factorint(da).items():
                if q % 2 and db % (p**k):
                    return

    if args.get('fast', True):
        try:
            result = field_isomorphism_pslq(a, b)

            if result is not None:
                return result
        except NotImplementedError:
            pass

    return field_isomorphism_factor(a, b)



# Tests for var are in their own file, because var pollutes global namespace.

import pytest

from diofant import Function, FunctionClass, Symbol, symbols, var
from diofant.abc import a, b, d, e, x, y


bb, cc, zz, _x, fg = symbols('bb cc zz _x fg')

__all__ = ()

# make z1 with call-depth = 1


def _make_z1():
    var('z1')

# make z2 with call-depth = 2


def __make_z2():
    var('z2')


def _make_z2():
    __make_z2()


def test_var():
    var('a')
    assert a == Symbol('a')

    var('b bb cc zz _x')
    assert b == Symbol('b')
    assert bb == Symbol('bb')
    assert cc == Symbol('cc')
    assert zz == Symbol('zz')
    assert _x == Symbol('_x')

    v = var(['d', 'e', 'fg'])
    assert d == Symbol('d')
    assert e == Symbol('e')
    assert fg == Symbol('fg')

    # check return value
    assert v == [d, e, fg]

    # pylint: disable=undefined-variable

    # see if var() really injects into global namespace
    pytest.raises(NameError, lambda: z1)  # type: ignore[name-defined]
    _make_z1()
    assert z1 == Symbol('z1')

    pytest.raises(NameError, lambda: z2)  # type: ignore[name-defined]
    _make_z2()
    assert z2 == Symbol('z2')


def test_var_return():
    pytest.raises(ValueError, lambda: var(''))
    v2 = var('q')
    v3 = var('q p')

    assert v2 == Symbol('q')
    assert v3 == (Symbol('q'), Symbol('p'))


def test_var_accepts_comma():
    v1 = var('x y z')
    v2 = var('x,y,z')
    v3 = var('x,y z')

    assert v1 == v2
    assert v1 == v3


def test_var_keywords():
    var('x y', extended_real=True)
    assert x.is_extended_real
    assert y.is_extended_real


def test_var_cls():
    f = var('f', cls=Function)

    assert isinstance(f, FunctionClass)

    g, h = var('g,h', cls=Function)

    assert isinstance(g, FunctionClass)
    assert isinstance(h, FunctionClass)


def test_var_nested():
    ((a, *_), (x, *_)) = var(('a:d', 'x:z'))
    assert isinstance(a, Symbol)
    assert isinstance(x, Symbol)



"""Generic Rules for Diofant.

This file assumes knowledge of Basic and little else.
"""

import functools

from ..utilities.iterables import sift
from .basic import Atom, Basic


__all__ = ('arguments', 'operator', 'term', 'rm_id',
           'glom', 'flatten', 'unpack', 'sort')


@functools.singledispatch
def arguments(o):
    """Extract arguments from an expression."""
    return o.args


@arguments.register(int)
@arguments.register(Atom)
def arguments_atomic(o):
    return ()


@functools.singledispatch
def operator(o):
    """Extract the head of an expression."""
    return o.func


@operator.register(int)
@operator.register(Atom)
def operator_atomic(o):
    return o


@functools.singledispatch
def term(op, args):
    """Build an expression from the head and arguments."""
    return op(*args)


@term.register(int)
@term.register(Atom)
def term_atomic(op, args):
    return op


# Functions that create rules


def rm_id(isid):
    """Create a rule to remove identities

    isid - fn :: x -> Bool  --- whether or not this element is an identity

    >>> remove_zeros = rm_id(lambda x: x == 0)
    >>> remove_zeros(Basic(1, 0, 2))
    Basic(1, 2)
    >>> remove_zeros(Basic(0, 0))  # If only identites then we keep one
    Basic(0)

    See Also
    ========

    unpack

    """
    def ident_remove(expr):
        """Remove identities."""
        ids = list(map(isid, arguments(expr)))
        if sum(ids) == 0:           # No identities. Common case
            return expr
        elif sum(ids) != len(ids):  # there is at least one non-identity
            return term(operator(expr),
                        [arg for arg, x in zip(arguments(expr), ids) if not x])
        else:
            return term(operator(expr), [arguments(expr)[0]])

    return ident_remove


def glom(key, count, combine):
    """Create a rule to conglomerate identical args.

    >>> def key(x):
    ...     return x.as_coeff_Mul()[1]
    >>> def count(x):
    ...     return x.as_coeff_Mul()[0]
    >>> def combine(cnt, arg):
    ...     return cnt * arg
    >>> rl = glom(key, count, combine)

    >>> rl(Add(x, -x, 3*x, 2, 3, evaluate=False))
    3*x + 5

    Wait, how are key, count and combine supposed to work?

    >>> key(2*x)
    x
    >>> count(2*x)
    2
    >>> combine(2, x)
    2*x

    """
    def conglomerate(expr):
        """Conglomerate together identical args x + x -> 2x."""
        groups = sift(arguments(expr), key)
        counts = {k: sum(map(count, args)) for k, args in groups.items()}
        newargs = [combine(cnt, mat) for mat, cnt in counts.items()]
        if set(newargs) != set(arguments(expr)):
            return term(operator(expr), newargs)
        else:
            return expr

    return conglomerate


def sort(key):
    """Create a rule to sort by a key function.

    >>> sort_rl = sort(str)
    >>> sort_rl(Basic(3, 1, 2))
    Basic(1, 2, 3)

    """

    def sort_rl(expr):
        return term(operator(expr), sorted(arguments(expr), key=key))
    return sort_rl


# Functions that are rules


def unpack(expr):
    """Rule to unpack singleton args.

    >>> unpack(Basic(2))
    2

    """
    if len(arguments(expr)) == 1:
        return arguments(expr)[0]
    else:
        return expr


def flatten(expr):
    """Flatten T(a, b, T(c, d), T2(e)) to T(a, b, c, d, T2(e))."""
    cls = operator(expr)
    args = []
    for arg in arguments(expr):
        if operator(arg) == cls:
            args.extend(arguments(arg))
        else:
            args.append(arg)
    return term(cls, args)


def identity(x):
    return x


def switch(key, ruledict):
    """Select a rule based on the result of key called on the function."""
    def switch_rl(expr):
        rl = ruledict.get(key(expr), identity)
        return rl(expr)
    return switch_rl


def typed(ruletypes):
    """Apply rules based on the expression type.

    Examples
    ========

    >>> rm_zeros = rm_id(lambda x: x == 0)
    >>> rm_ones = rm_id(lambda x: x == 1)
    >>> remove_idents = typed({Add: rm_zeros, Mul: rm_ones})
    """
    return switch(type, ruletypes)


def treeapply(tree, join, leaf=identity):
    """Apply functions onto recursive containers (tree).

    join - a dictionary mapping container types to functions
      e.g. ``{list: minimize, tuple: chain}``

    Keys are containers/iterables.  Values are functions [a] -> a.

    Examples
    ========

    >>> tree = [(3, 2), (4, 1)]
    >>> treeapply(tree, {list: max, tuple: min})
    2

    >>> def mul(*args):
    ...     total = 1
    ...     for arg in args:
    ...         total *= arg
    ...     return total
    >>> treeapply(tree, {list: mul, tuple: lambda *args: sum(args)})
    25
    """
    for typ in join:
        if isinstance(tree, typ):
            return join[typ](*map(functools.partial(treeapply, join=join, leaf=leaf),
                                  tree))
    return leaf(tree)


def minimize(*rules, objective=identity):
    """Select result of rules that minimizes objective.

    Examples
    ========

    >>> from diofant.core.strategies import minimize

    >>> rl = minimize(lambda x: x + 1, lambda x: x - 1)
    >>> rl(4)
    3
    """
    def minrule(expr):
        return min((rule(expr) for rule in rules), key=objective)
    return minrule


def chain(*rules):
    """Compose a sequence of rules so that they apply to the expr sequentially."""
    def chain_rl(expr):
        for rule in rules:
            expr = rule(expr)
        return expr
    return chain_rl


def greedy(tree, objective=identity, **kwargs):
    """Execute a strategic tree.  Select alternatives greedily,

    Examples
    ========

    >>> tree = [lambda x: x + 1,
    ...         (lambda x: x - 1, lambda x: 2*x)]  # either inc or dec-then-double
    >>> fn = greedy(tree)
    >>> fn(4)  # lowest value comes from the inc
    5
    >>> fn(1)  # lowest value comes from dec then double
    0

    This function selects between options in a tuple.  The result is chosen that
    minimizes the objective function.

    >>> fn = greedy(tree, objective=lambda x: -x)  # maximize
    >>> fn(4)  # highest value comes from the dec then double
    6
    >>> fn(1)  # highest value comes from the inc
    2
    """
    optimize = functools.partial(minimize, objective=objective)
    return treeapply(tree, {list: optimize, tuple: chain}, **kwargs)


def do_one(rules):
    """Try each of the rules until one works. Then stop."""
    def do_one_rl(expr):
        for rl in rules:
            result = rl(expr)
            if result != expr:
                return result
        return expr
    return do_one_rl


def condition(cond, rule):
    """Only apply rule if condition is true."""
    def conditioned_rl(expr):
        if cond(expr):
            return rule(expr)
        else:
            return expr
    return conditioned_rl


def exhaust(rule):
    """Apply a rule repeatedly until it has no effect."""
    def exhaustive_rl(expr):
        new, old = rule(expr), expr
        while new != old:
            new, old = rule(new), new
        return new
    return exhaustive_rl


basic_fns = {'op': type,
             'new': Basic.__new__,
             'leaf': lambda x: not isinstance(x, Basic) or x.is_Atom,
             'children': lambda x: x.args}


def sall(rule, fns=basic_fns):
    """Strategic all - apply rule to args."""
    op, new, children, leaf = map(fns.get, ('op', 'new', 'children', 'leaf'))

    def all_rl(expr):
        if leaf(expr):
            return expr
        else:
            args = map(rule, children(expr))
            return new(op(expr), *args)

    return all_rl


def bottom_up(rule, fns=basic_fns):
    """Apply a rule down a tree running it on the bottom nodes first."""
    def rec(expr):
        return sall(bottom_up(rule, fns), fns)(expr)
    return chain(rec, rule)


def null_safe(rule):
    """Return original expr if rule returns None."""
    def null_safe_rl(expr):
        result = rule(expr)
        if result is None:
            return expr
        else:
            return result
    return null_safe_rl



"""Diofant is a Python library for symbolic mathematics."""

from fractions import Fraction
import pkg_resources
__version__ = pkg_resources.get_distribution(__name__).version
del pkg_resources

from .core import (Add, Atom, AtomicExpr, Basic, Catalan, Derivative, Dict,
                   Dummy, E, Eq, Equality, EulerGamma, Expr, Float, Function,
                   FunctionClass, Ge, GoldenRatio, GreaterThan, Gt, I, Integer,
                   Lambda, Le, LessThan, Lt, Mod, Mul, N, Ne, Number,
                   NumberSymbol, PoleError, Pow, PrecisionExhausted, Rational,
                   Rel, Relational, S, StrictGreaterThan, StrictLessThan, Subs,
                   Symbol, SympifyError, Tuple, Unequality, Wild, WildFunction,
                   cacheit, comp, count_ops, diff, evaluate, expand,
                   expand_complex, expand_func, expand_log, expand_mul,
                   expand_multinomial, expand_power_base, expand_power_exp,
                   expand_trig, factor_nc, factor_terms, gcd_terms,
                   integer_digits, integer_nthroot, mod_inverse, nan, nfloat,
                   oo, pi, preorder_traversal, symbols, sympify, var,
                   vectorize, zoo)
from .logic import (ITE, And, Equivalent, Implies, Nand, Nor, Not, Or,
                    Xor, false, satisfiable, simplify_logic, to_cnf,
                    to_dnf, to_nnf, true)
from .polys import (LC, LM, LT, BasePolynomialError, CoercionFailed,
                    ComputationFailed, DomainError, EvaluationFailed,
                    ExactQuotientFailed, ExtraneousFactors, FlagError,
                    FractionField, GeneratorsError, GeneratorsNeeded,
                    GroebnerBasis, HeuristicGCDFailed, HomomorphismFailed,
                    IsomorphismFailed, Monomial, MultivariatePolynomialError,
                    NotAlgebraic, NotInvertible, NotReversible,
                    OperationNotSupported, OptionError, Options,
                    PolificationFailed, Poly, PolynomialDivisionFailed,
                    PolynomialError, PolynomialRing, PurePoly,
                    RefinementFailed, RootOf, RootSum, UnificationFailed,
                    UnivariatePolynomialError, UnivarPolynomialRing, apart,
                    apart_list, assemble_partfrac_list, cancel,
                    chebyshevt_poly, chebyshevu_poly, cofactors, compose,
                    construct_domain, content, count_roots, cyclotomic_poly,
                    decompose, degree, discriminant, div, exquo,
                    factor, factor_list, field, field_isomorphism, gcd,
                    gcdex, grevlex, grlex, groebner, half_gcdex,
                    hermite_poly, horner, igrevlex, igrlex, ilex, interpolate,
                    interpolating_poly, invert, itermonomials, jacobi_poly,
                    laguerre_poly, lcm, legendre_poly, lex,
                    minimal_polynomial, monic, nroots, parallel_poly_from_expr,
                    primitive, primitive_element, quo, random_poly,
                    real_roots, reduced, rem, resultant, ring, roots,
                    spherical_bessel_fn, sqf, sqf_list, sqf_norm, sqf_part,
                    subresultants, swinnerton_dyer_poly, symmetric_poly,
                    symmetrize, terms_gcd, together, trunc, viete)
from .domains import (CC, EX, FF, GF, GROUND_TYPES, QQ, RR, ZZ, AlgebraicField,
                      ComplexAlgebraicField, ComplexField, Domain,
                      ExpressionDomain, FF_gmpy, FF_python, FiniteField,
                      IntegerRing, PythonRational, QQ_gmpy, QQ_python,
                      RationalField, RealAlgebraicField, RealField, ZZ_gmpy,
                      ZZ_python, IntegerModRing)
from .series import Limit, O, Order, limit, residue
from .functions import (E1, Abs, Chi, Ci, DiracDelta, Ei, Eijk,
                        FallingFactorial, Heaviside, Id, KroneckerDelta,
                        LambertW, LeviCivita, Li, Max, Min, Piecewise,
                        RisingFactorial, Shi, Si, Ynm, Ynm_c, Znm, acos, acosh,
                        acot, acoth, acsc, adjoint, airyai, airyaiprime,
                        airybi, airybiprime, arg, asec, asin, asinh,
                        assoc_laguerre, assoc_legendre, atan, atan2, atanh,
                        bell, bernoulli, besseli, besselj, besselk, bessely,
                        beta, binomial, bspline_basis, bspline_basis_set,
                        catalan, cbrt, ceiling, chebyshevt, chebyshevt_root,
                        chebyshevu, chebyshevu_root, conjugate, cos, cosh, cot,
                        coth, csc, csch, digamma, dirichlet_eta, elliptic_e,
                        elliptic_f, elliptic_k, elliptic_pi, erf, erf2,
                        erf2inv, erfc, erfcinv, erfi, erfinv, euler, exp,
                        exp_polar, expint, factorial, factorial2, ff,
                        fibonacci, floor, fresnelc, fresnels, gamma,
                        gegenbauer, genocchi, hankel1, hankel2, harmonic,
                        hermite, hyper, im, jacobi, jacobi_normalized, jn,
                        jn_zeros, laguerre, legendre, lerchphi, li, ln, log,
                        loggamma, lowergamma, lucas, meijerg,
                        periodic_argument, piecewise_fold, polar_lift,
                        polarify, polygamma, polylog, principal_branch, re,
                        real_root, rf, root, sec, sech, sign, sin, sinh, sqrt,
                        subfactorial, tan, tanh, transpose, trigamma,
                        unbranched_argument, unpolarify, uppergamma, yn, zeta)
from .ntheory import (Sieve, binomial_coefficients, binomial_coefficients_list,
                      continued_fraction_convergents,
                      continued_fraction_iterator, continued_fraction_periodic,
                      continued_fraction_reduce, cycle_length, discrete_log,
                      divisor_count, divisor_sigma, divisors,
                      egyptian_fraction, factorint, factorrat,
                      is_nthpow_residue, is_primitive_root, is_quad_residue,
                      is_square, isprime, jacobi_symbol, legendre_symbol,
                      mobius, multinomial_coefficients, multiplicity, n_order,
                      nextprime, npartitions, nthroot_mod, perfect_power,
                      pollard_pm1, pollard_rho, prevprime, prime, primefactors,
                      primepi, primerange, primitive_root, primorial,
                      quadratic_residues, randprime, sieve, sqrt_mod,
                      sqrt_mod_iter, square_factor, totient, trailing)
from .concrete import Product, Sum, product, summation
from .simplify import (FU, EPath, besselsimp, bottom_up, collect,
                       collect_const, combsimp, cse, denom, epath, exptrigsimp,
                       fraction, fu, hyperexpand, hypersimilar, hypersimp,
                       logcombine, nsimplify, numer, posify, powdenest,
                       powsimp, radsimp, ratsimp, ratsimpmodprime, rcollect,
                       separatevars, signsimp, simplify, sqrtdenest, trigsimp,
                       use)
from .sets import (Complement, Contains, EmptySet, FiniteSet, ImageSet,
                   Integers, Intersection, Interval, Naturals, Naturals0,
                   ProductSet, Range, Rationals, Reals, Set,
                   SymmetricDifference, Union, imageset, ExtendedReals)
from .solvers import dsolve, reduce_inequalities, rsolve, solve
from .matrices import (Adjoint, BlockDiagMatrix, BlockMatrix, Determinant,
                       DiagonalMatrix, DiagonalOf, FunctionMatrix, GramSchmidt,
                       HadamardProduct, Identity, ImmutableDenseMatrix,
                       ImmutableMatrix, ImmutableSparseMatrix, Inverse, MatAdd,
                       MatMul, MatPow, Matrix, MatrixBase, MatrixExpr,
                       MatrixSlice, MatrixSymbol, MutableDenseMatrix,
                       MutableMatrix, MutableSparseMatrix,
                       NonSquareMatrixError, ShapeError, SparseMatrix, Trace,
                       Transpose, ZeroMatrix, block_collapse, blockcut,
                       casoratian, det, diag, eye, hadamard_product, hessian,
                       jordan_cell, list2numpy, matrix2numpy,
                       matrix_multiply_elementwise, ones, randMatrix,
                       rot_axis1, rot_axis2, rot_axis3, symarray, trace,
                       vandermonde, wronskian, zeros)
from .geometry import (Circle, Curve, Ellipse, GeometryError, Line,
                       Point, Polygon, Ray,
                       RegularPolygon, Segment, Triangle,
                       are_similar, centroid, convex_hull, deg, idiff,
                       intersection, rad)
from .utilities import (cantor_product, capture, default_sort_key, filldedent,
                        flatten, group, has_dups, has_variety, lambdify,
                        numbered_symbols, ordered, postfixes,
                        postorder_traversal, prefixes, sift, subsets,
                        topological_sort, unflatten, variations)
from .integrals import (CosineTransform, FourierTransform, HankelTransform,
                        Integral, InverseCosineTransform,
                        InverseFourierTransform, InverseHankelTransform,
                        InverseLaplaceTransform, InverseMellinTransform,
                        InverseSineTransform, LaplaceTransform,
                        MellinTransform, SineTransform, cosine_transform,
                        fourier_transform, hankel_transform, integrate,
                        inverse_cosine_transform, inverse_fourier_transform,
                        inverse_hankel_transform, inverse_laplace_transform,
                        inverse_mellin_transform, inverse_sine_transform,
                        laplace_transform, line_integrate, mellin_transform,
                        sine_transform)
from .tensor import (Array, DenseNDimArray, Idx, ImmutableDenseNDimArray,
                     ImmutableSparseNDimArray, Indexed, IndexedBase,
                     MutableDenseNDimArray, MutableSparseNDimArray, NDimArray,
                     SparseNDimArray, derive_by_array,
                     get_contraction_structure, get_indices, permutedims,
                     tensorcontraction, tensorproduct)
from .calculus import maximize, minimize
from .combinatorics import (AbelianGroup, AlternatingGroup, Cycle, CyclicGroup,
                            DihedralGroup, DirectProduct, GrayCode,
                            IntegerPartition, Partition, Permutation,
                            PermutationGroup, Polyhedron, Prufer, RGS_enum,
                            RGS_rank, RGS_unrank, RubikGroup, Subset,
                            SymmetricGroup, alternating, cube, cyclic,
                            dihedral, dodecahedron, icosahedron, octahedron,
                            symmetric, tetrahedron)
from .plotting import (plot, plot3d, plot3d_parametric_line,
                       plot3d_parametric_surface, plot_backends, plot_implicit,
                       plot_parametric)
from .printing import (StrPrinter, ccode, dotprint, fcode, latex,
                       mathematica_code, mathml, octave_code, pprint,
                       pprint_use_unicode, pretty, pretty_print, python, srepr,
                       sstr, sstrrepr)
from .interactive import init_printing


__all__ = (
    'Add', 'Atom', 'AtomicExpr', 'Basic', 'Catalan', 'Derivative', 'Dict',
    'Dummy', 'E', 'Eq', 'Equality', 'EulerGamma', 'Expr', 'Float', 'Function',
    'FunctionClass', 'Ge', 'GoldenRatio', 'GreaterThan', 'Gt', 'I', 'Integer',
    'Lambda', 'Le', 'LessThan', 'Lt', 'Mod', 'Mul', 'N', 'Ne', 'Number',
    'NumberSymbol', 'PoleError', 'Pow', 'PrecisionExhausted', 'Rational',
    'Rel', 'Relational', 'S', 'StrictGreaterThan', 'StrictLessThan', 'Subs',
    'Symbol', 'SympifyError', 'Tuple', 'Unequality', 'Wild', 'WildFunction',
    'cacheit', 'comp', 'count_ops', 'diff', 'evaluate', 'expand',
    'expand_complex', 'expand_func', 'expand_log', 'expand_mul',
    'expand_multinomial', 'expand_power_base', 'expand_power_exp',
    'expand_trig', 'factor_nc', 'factor_terms', 'gcd_terms',
    'integer_digits', 'integer_nthroot', 'mod_inverse', 'nan', 'nfloat',
    'oo', 'pi', 'preorder_traversal', 'symbols', 'sympify', 'var',
    'vectorize', 'zoo', 'ITE', 'And', 'Equivalent', 'Implies', 'Nand', 'Nor',
    'Not', 'Or', 'Xor', 'false',
    'satisfiable', 'simplify_logic', 'to_cnf', 'to_dnf', 'to_nnf', 'true',
    'LC', 'LM', 'LT', 'BasePolynomialError', 'CoercionFailed',
    'ComputationFailed', 'DomainError', 'EvaluationFailed',
    'ExactQuotientFailed', 'ExtraneousFactors', 'FlagError',
    'FractionField', 'GeneratorsError', 'GeneratorsNeeded',
    'GroebnerBasis', 'HeuristicGCDFailed', 'HomomorphismFailed',
    'IsomorphismFailed', 'Monomial', 'MultivariatePolynomialError',
    'NotAlgebraic', 'NotInvertible', 'NotReversible', 'OperationNotSupported',
    'OptionError', 'Options', 'PolificationFailed', 'Poly',
    'PolynomialDivisionFailed', 'PolynomialError', 'PolynomialRing', 'PurePoly',
    'RefinementFailed', 'RootOf', 'RootSum', 'UnificationFailed',
    'UnivariatePolynomialError', 'UnivarPolynomialRing', 'apart',
    'apart_list', 'assemble_partfrac_list', 'cancel', 'chebyshevt_poly',
    'chebyshevu_poly', 'cofactors', 'compose', 'construct_domain', 'content',
    'count_roots', 'cyclotomic_poly', 'decompose', 'degree',
    'discriminant', 'div', 'exquo', 'factor', 'factor_list', 'field',
    'field_isomorphism', 'gcd', 'gcdex', 'grevlex', 'grlex',
    'groebner', 'half_gcdex', 'hermite_poly', 'horner', 'igrevlex', 'igrlex',
    'ilex', 'interpolate', 'interpolating_poly', 'invert', 'itermonomials',
    'jacobi_poly', 'laguerre_poly', 'lcm', 'legendre_poly', 'lex',
    'minimal_polynomial', 'monic', 'nroots', 'parallel_poly_from_expr',
    'primitive', 'primitive_element', 'quo', 'random_poly',
    'real_roots', 'reduced', 'rem', 'resultant', 'ring', 'roots',
    'spherical_bessel_fn', 'sqf', 'sqf_list', 'sqf_norm', 'sqf_part',
    'subresultants', 'swinnerton_dyer_poly', 'symmetric_poly',
    'symmetrize', 'terms_gcd', 'together', 'trunc', 'viete',
    'CC', 'EX', 'FF', 'GF', 'GROUND_TYPES', 'QQ', 'RR', 'ZZ', 'AlgebraicField',
    'ComplexAlgebraicField', 'ComplexField', 'Domain', 'ExpressionDomain',
    'FF_gmpy', 'FF_python', 'FiniteField', 'IntegerRing', 'PythonRational',
    'QQ_gmpy', 'QQ_python', 'RationalField', 'RealAlgebraicField', 'RealField',
    'ZZ_gmpy', 'ZZ_python', 'Limit', 'O', 'Order', 'limit', 'residue',
    'E1', 'Abs', 'Chi', 'Ci', 'DiracDelta', 'Ei', 'Eijk', 'FallingFactorial',
    'Heaviside', 'Id', 'KroneckerDelta', 'LambertW', 'LeviCivita', 'Li', 'Max',
    'Min', 'Piecewise', 'RisingFactorial', 'Shi', 'Si', 'Ynm', 'Ynm_c', 'Znm',
    'acos', 'acosh', 'acot', 'acoth', 'acsc', 'adjoint', 'airyai',
    'airyaiprime', 'airybi', 'airybiprime', 'arg', 'asec', 'asin', 'asinh',
    'assoc_laguerre', 'assoc_legendre', 'atan', 'atan2', 'atanh',
    'bell', 'bernoulli', 'besseli', 'besselj', 'besselk', 'bessely',
    'beta', 'binomial', 'bspline_basis', 'bspline_basis_set', 'catalan',
    'cbrt', 'ceiling', 'chebyshevt', 'chebyshevt_root', 'chebyshevu',
    'chebyshevu_root', 'conjugate', 'cos', 'cosh', 'cot', 'coth', 'csc',
    'csch', 'digamma', 'dirichlet_eta', 'elliptic_e', 'elliptic_f',
    'elliptic_k', 'elliptic_pi', 'erf', 'erf2', 'erf2inv', 'erfc', 'erfcinv',
    'erfi', 'erfinv', 'euler', 'exp', 'exp_polar', 'expint', 'factorial',
    'factorial2', 'ff', 'fibonacci', 'floor', 'fresnelc', 'fresnels', 'gamma',
    'gegenbauer', 'genocchi', 'hankel1', 'hankel2', 'harmonic', 'hermite',
    'hyper', 'im', 'jacobi', 'jacobi_normalized', 'jn', 'jn_zeros', 'laguerre',
    'legendre', 'lerchphi', 'li', 'ln', 'log', 'loggamma', 'lowergamma',
    'lucas', 'meijerg', 'periodic_argument', 'piecewise_fold', 'polar_lift',
    'polarify', 'polygamma', 'polylog', 'principal_branch', 're', 'real_root',
    'rf', 'root', 'sec', 'sech', 'sign', 'sin', 'sinh', 'sqrt', 'subfactorial',
    'tan', 'tanh', 'transpose', 'trigamma', 'unbranched_argument', 'unpolarify',
    'uppergamma', 'yn', 'zeta', 'Sieve', 'binomial_coefficients',
    'binomial_coefficients_list', 'continued_fraction_convergents',
    'continued_fraction_iterator', 'continued_fraction_periodic',
    'continued_fraction_reduce', 'cycle_length', 'discrete_log',
    'divisor_count', 'divisor_sigma', 'divisors', 'egyptian_fraction',
    'factorint', 'factorrat', 'is_nthpow_residue', 'is_primitive_root',
    'is_quad_residue', 'is_square', 'isprime', 'jacobi_symbol',
    'legendre_symbol', 'mobius', 'multinomial_coefficients', 'multiplicity',
    'n_order', 'nextprime', 'npartitions', 'nthroot_mod', 'perfect_power',
    'pollard_pm1', 'pollard_rho', 'prevprime', 'prime', 'primefactors',
    'primepi', 'primerange', 'primitive_root', 'primorial',
    'quadratic_residues', 'randprime', 'sieve', 'sqrt_mod', 'sqrt_mod_iter',
    'square_factor', 'totient', 'trailing', 'Product', 'Sum', 'product',
    'summation', 'FU', 'EPath', 'besselsimp', 'bottom_up', 'collect',
    'collect_const', 'combsimp', 'cse', 'denom', 'epath', 'exptrigsimp',
    'fraction', 'fu', 'hyperexpand', 'hypersimilar', 'hypersimp',
    'logcombine', 'nsimplify', 'numer', 'posify', 'powdenest', 'powsimp',
    'radsimp', 'ratsimp', 'ratsimpmodprime', 'rcollect', 'separatevars',
    'signsimp', 'simplify', 'sqrtdenest', 'trigsimp', 'use', 'Complement',
    'Contains', 'EmptySet', 'FiniteSet', 'ImageSet', 'Integers',
    'Intersection', 'Interval', 'Naturals', 'Naturals0', 'ProductSet',
    'Range', 'Rationals', 'Reals', 'Set', 'SymmetricDifference', 'Union',
    'imageset', 'dsolve', 'reduce_inequalities', 'rsolve', 'solve',
    'Adjoint', 'BlockDiagMatrix', 'BlockMatrix', 'Determinant',
    'DiagonalMatrix', 'DiagonalOf', 'FunctionMatrix', 'GramSchmidt',
    'HadamardProduct', 'Identity', 'ImmutableDenseMatrix',
    'ImmutableMatrix', 'ImmutableSparseMatrix', 'Inverse', 'MatAdd',
    'MatMul', 'MatPow', 'Matrix', 'MatrixBase', 'MatrixExpr',
    'MatrixSlice', 'MatrixSymbol', 'MutableDenseMatrix', 'MutableMatrix',
    'MutableSparseMatrix', 'NonSquareMatrixError', 'ShapeError', 'SparseMatrix',
    'Trace', 'Transpose', 'ZeroMatrix', 'block_collapse', 'blockcut',
    'casoratian', 'det', 'diag', 'eye', 'hadamard_product', 'hessian',
    'jordan_cell', 'list2numpy', 'matrix2numpy', 'matrix_multiply_elementwise',
    'ones', 'randMatrix', 'rot_axis1', 'rot_axis2', 'rot_axis3', 'symarray',
    'trace', 'vandermonde', 'wronskian', 'zeros', 'Circle', 'Curve', 'Ellipse',
    'GeometryError', 'Line', 'Point', 'Fraction',
    'Polygon', 'Ray', 'RegularPolygon', 'Segment',
    'Triangle', 'are_similar', 'centroid', 'convex_hull', 'deg', 'idiff',
    'intersection', 'rad', 'cantor_product', 'capture', 'default_sort_key',
    'filldedent', 'flatten', 'group', 'has_dups', 'has_variety', 'lambdify',
    'numbered_symbols', 'ordered', 'postfixes', 'postorder_traversal',
    'prefixes', 'sift', 'subsets', 'topological_sort', 'unflatten',
    'variations', 'CosineTransform', 'FourierTransform', 'HankelTransform',
    'Integral', 'InverseCosineTransform', 'InverseFourierTransform',
    'InverseHankelTransform', 'InverseLaplaceTransform',
    'InverseMellinTransform', 'InverseSineTransform', 'LaplaceTransform',
    'MellinTransform', 'SineTransform', 'cosine_transform', 'fourier_transform',
    'hankel_transform', 'integrate', 'inverse_cosine_transform',
    'inverse_fourier_transform', 'inverse_hankel_transform',
    'inverse_laplace_transform', 'inverse_mellin_transform',
    'inverse_sine_transform', 'laplace_transform', 'line_integrate',
    'mellin_transform', 'sine_transform', 'Array', 'DenseNDimArray', 'Idx',
    'ImmutableDenseNDimArray', 'ImmutableSparseNDimArray', 'Indexed',
    'IndexedBase', 'MutableDenseNDimArray', 'MutableSparseNDimArray',
    'NDimArray', 'SparseNDimArray', 'derive_by_array',
    'get_contraction_structure', 'get_indices', 'permutedims',
    'tensorcontraction', 'tensorproduct', 'maximize', 'minimize',
    'AbelianGroup', 'AlternatingGroup', 'Cycle', 'CyclicGroup',
    'DihedralGroup', 'DirectProduct', 'GrayCode', 'IntegerPartition',
    'Partition', 'Permutation', 'PermutationGroup', 'Polyhedron', 'Prufer',
    'RGS_enum', 'RGS_rank', 'RGS_unrank', 'RubikGroup', 'Subset',
    'SymmetricGroup', 'alternating', 'cube', 'cyclic', 'dihedral',
    'dodecahedron', 'icosahedron', 'octahedron', 'symmetric', 'tetrahedron',
    'plot', 'plot3d', 'plot3d_parametric_line', 'plot3d_parametric_surface',
    'plot_backends', 'plot_implicit', 'plot_parametric', 'StrPrinter', 'ccode',
    'dotprint', 'fcode', 'latex', 'mathematica_code', 'mathml', 'octave_code',
    'pprint', 'pprint_use_unicode', 'pretty', 'pretty_print', 'python', 'srepr',
    'sstr', 'sstrrepr', 'init_printing', 'ExtendedReals', 'IntegerModRing')



# If the arbitrary constant class from issue sympy/sympy#4435 is ever
# implemented, this should serve as a set of test cases.

import pytest

from diofant import (Eq, Function, I, Integer, Integral, Pow, Symbol, acos,
                     cos, cosh, dsolve, exp, log, sin, sinh, sqrt, symbols)
from diofant.abc import x, y, z
from diofant.solvers.ode import constant_renumber, constantsimp


__all__ = ()

C1, C2, C3 = symbols('C1:4')
f = Function('f')


def test_constant_mul():
    # We want C1 (Constant) below to absorb the y's, but not the x's
    assert constant_renumber(constantsimp(y*C1, [C1]), 'C', 1, 1) == C1*y
    assert constant_renumber(constantsimp(C1*y, [C1]), 'C', 1, 1) == C1*y
    assert constant_renumber(constantsimp(x*C1, [C1]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(C1*x, [C1]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(2*C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1*2, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(y*C1*x, [C1, y]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp(x*y*C1, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(y*x*C1, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(C1*x*y, [C1, y]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp(x*C1*y, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(C1*y*(y + 1), [C1]), 'C', 1, 1) == C1*y*(y+1)
    assert constant_renumber(constantsimp(y*C1*(y + 1), [C1]), 'C', 1, 1) == C1*y*(y+1)
    assert constant_renumber(constantsimp(x*(y*C1), [C1]), 'C', 1, 1) == x*y*C1
    assert constant_renumber(constantsimp(x*(C1*y), [C1]), 'C', 1, 1) == x*y*C1
    assert constant_renumber(constantsimp(C1*(x*y), [C1, y]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp((x*y)*C1, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp((y*x)*C1, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp(y*(y + 1)*C1, [C1, y]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp((C1*x)*y, [C1, y]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp(y*(x*C1), [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(constantsimp((x*C1)*y, [C1, y]), 'C', 1, 1) == x*C1
    assert constant_renumber(
        constantsimp(C1*x*y*x*y*2, [C1, y]), 'C', 1, 1) == C1*x**2
    assert constant_renumber(constantsimp(C1*x*y*z, [C1, y, z]), 'C', 1, 1) == C1*x
    assert constant_renumber(
        constantsimp(C1*x*y**2*sin(z), [C1, y, z]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp(C1*C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1*C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C2*C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C1*C1*C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(
        constantsimp(C1*x*2**x, [C1]), 'C', 1, 1) == C1*x*2**x


def test_constant_add():
    assert constant_renumber(constantsimp(C1 + C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1 + 2, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(2 + C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1 + y, [C1, y]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1 + x, [C1]), 'C', 1, 1) == C1 + x
    assert constant_renumber(constantsimp(C1 + C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1 + C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C2 + C1, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C1 + C2 + C1, [C1, C2]), 'C', 1, 2) == C1


def test_constant_power_as_base():
    assert constant_renumber(constantsimp(C1**C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(Pow(C1, C1), [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1**C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1**C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C2**C1, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C2**C2, [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(C1**y, [C1, y]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(C1**x, [C1]), 'C', 1, 1) == C1**x
    assert constant_renumber(constantsimp(C1**2, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(
        constantsimp(C1**(x*y), [C1]), 'C', 1, 1) == C1**(x*y)


def test_constant_power_as_exp():
    assert constant_renumber(constantsimp(x**C1, [C1]), 'C', 1, 1) == x**C1
    assert constant_renumber(constantsimp(y**C1, [C1, y]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(x**y**C1, [C1, y]), 'C', 1, 1) == x**C1
    assert constant_renumber(
        constantsimp((x**y)**C1, [C1]), 'C', 1, 1) == (x**y)**C1
    assert constant_renumber(
        constantsimp(x**(y**C1), [C1, y]), 'C', 1, 1) == x**C1
    assert constant_renumber(constantsimp(x**C1**y, [C1, y]), 'C', 1, 1) == x**C1
    assert constant_renumber(
        constantsimp(x**(C1**y), [C1, y]), 'C', 1, 1) == x**C1
    assert constant_renumber(
        constantsimp((x**C1)**y, [C1]), 'C', 1, 1) == (x**C1)**y
    assert constant_renumber(constantsimp(2**C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(Integer(2)**C1, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(exp(C1), [C1]), 'C', 1, 1) == C1
    assert constant_renumber(
        constantsimp(exp(C1 + x), [C1]), 'C', 1, 1) == C1*exp(x)
    assert constant_renumber(constantsimp(Pow(2, C1), [C1]), 'C', 1, 1) == C1


def test_constant_function():
    assert constant_renumber(constantsimp(sin(C1), [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(f(C1), [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(f(C1, C1), [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(f(C1, C2), [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(f(C2, C1), [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(f(C2, C2), [C1, C2]), 'C', 1, 2) == C1
    assert constant_renumber(
        constantsimp(f(C1, x), [C1]), 'C', 1, 2) == f(C1, x)
    assert constant_renumber(constantsimp(f(C1, y), [C1, y]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(f(y, C1), [C1, y]), 'C', 1, 2) == C1
    assert constant_renumber(constantsimp(f(C1, y, C2), [C1, C2, y]), 'C', 1, 2) == C1


def test_constant_function_multiple():
    # The rules to not renumber in this case would be too complicated, and
    # dsolve is not likely to ever encounter anything remotely like this.
    assert constant_renumber(
        constantsimp(f(C1, C1, x), [C1]), 'C', 1, 1) == f(C1, C1, x)


def test_constant_multiple():
    assert constant_renumber(constantsimp(C1*2 + 2, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(constantsimp(x*2/C1, [C1]), 'C', 1, 1) == C1*x
    assert constant_renumber(constantsimp(C1**2*2 + 2, [C1]), 'C', 1, 1) == C1
    assert constant_renumber(
        constantsimp(sin(2*C1) + x + sqrt(2), [C1]), 'C', 1, 1) == C1 + x
    assert constant_renumber(constantsimp(2*C1 + C2, [C1, C2]), 'C', 1, 2) == C1


def test_constant_repeated():
    assert C1 + C1*x == constant_renumber(C1 + C1*x, 'C', 1, 3)


def test_ode_solutions():
    # only a few examples here, the rest will be tested in the actual dsolve tests
    assert constant_renumber(constantsimp(C1*exp(2*x) + exp(x)*(C2 + C3), [C1, C2, C3]), 'C', 1, 3) == \
        constant_renumber((C1*exp(x) + C2*exp(2*x)), 'C', 1, 2)
    assert constant_renumber(
        constantsimp(Eq(f(x), I*C1*sinh(x/3) + C2*cosh(x/3)), [C1, C2]),
        'C', 1, 2) == constant_renumber(Eq(f(x), C1*sinh(x/3) + C2*cosh(x/3)), 'C', 1, 2)
    assert constant_renumber(constantsimp(Eq(f(x), acos((-C1)/cos(x))), [C1]), 'C', 1, 1) == \
        Eq(f(x), acos(C1/cos(x)))
    assert constant_renumber(
        constantsimp(Eq(log(f(x)/C1) + 2*exp(x/f(x)), 0), [C1]),
        'C', 1, 1) == Eq(log(C1*f(x)) + 2*exp(x/f(x)), 0)
    assert constant_renumber(constantsimp(Eq(log(x*sqrt(2)*sqrt(1/x)*sqrt(f(x))
                                                 / C1) + x**2/(2*f(x)**2), 0), [C1]), 'C', 1, 1) == \
        Eq(log(C1*sqrt(x)*sqrt(f(x))) + x**2/(2*f(x)**2), 0)
    assert constant_renumber(constantsimp(Eq(-exp(-f(x)/x)*sin(f(x)/x)/2 + log(x/C1) -
                                             cos(f(x)/x)*exp(-f(x)/x)/2, 0), [C1]), 'C', 1, 1) == \
        Eq(-exp(-f(x)/x)*sin(f(x)/x)/2 + log(C1*x) - cos(f(x)/x) *
           exp(-f(x)/x)/2, 0)
    u2 = Symbol('u2')
    _a = Symbol('_a')
    assert constant_renumber(constantsimp(Eq(-Integral(-1/(sqrt(1 - u2**2)*u2),
                                                       (u2, _a, x/f(x))) + log(f(x)/C1), 0), [C1]), 'C', 1, 1) == \
        Eq(-Integral(-1/(u2*sqrt(1 - u2**2)), (u2, _a, x/f(x))) +
           log(C1*f(x)), 0)
    assert [constantsimp(i, [C1]) for i in [Eq(f(x), sqrt(-C1*x + x**2)), Eq(f(x), -sqrt(-C1*x + x**2))]] == \
        [Eq(f(x), sqrt(x*(C1 + x))), Eq(f(x), -sqrt(x*(C1 + x)))]

    # issue sympy/sympy5770
    k = Symbol('k', extended_real=True)
    t = Symbol('t')
    w = Function('w')
    sol = dsolve(w(t).diff((t, 6)) - k**6*w(t), w(t))
    assert len([s for s in sol.free_symbols if s.name.startswith('C')]) == 6
    assert constantsimp((C1*cos(x) + C2*cos(x))*exp(x), {C1, C2}) == \
        C1*cos(x)*exp(x)
    assert constantsimp(C1*cos(x) + C2*cos(x) + C3*sin(x), {C1, C2, C3}) == \
        C1*cos(x) + C3*sin(x)
    assert constantsimp(exp(C1 + x), {C1}) == C1*exp(x)
    assert constantsimp(x + C1 + y, {C1, y}) == C1 + x
    assert constantsimp(x + C1 + Integral(x, (x, 1, 2)), {C1}) == C1 + x


@pytest.mark.xfail
def test_nonlocal_simplification():
    assert constantsimp(C1 + C2+x*C2, [C1, C2]) == C1 + C2*x


def test_constant_Eq():
    # C1 on the rhs is well-tested, but the lhs is only tested here
    assert constantsimp(Eq(C1, 3 + f(x)*x), [C1]) == Eq(x*f(x), C1)
    assert constantsimp(Eq(C1, 3 * f(x)*x), [C1]) == Eq(f(x)*x, C1)



"""
Reimplementations of constructs introduced in later versions of Python than
we support. Also some functions that are needed Diofant-wide and are located
here for easy import.
"""

import os
import typing

from ..external import import_module


# These are in here because telling if something is an iterable just by calling
# hasattr(obj, "__iter__") behaves differently in Python 2 and Python 3.  In
# particular, hasattr(str, "__iter__") is False in Python 2 and True in Python 3.
# I think putting them here also makes it easier to use them in the core.

class NotIterable:
    """
    Use this as mixin when creating a class which is not supposed to return
    true when iterable() is called on its instances. I.e. avoid infinite loop
    when calling e.g. list() on the instance

    """


def iterable(i, exclude=(str, dict, NotIterable)):
    """
    Return a boolean indicating whether ``i`` is Diofant iterable.
    True also indicates that the iterator is finite, i.e. you e.g.
    call list(...) on the instance.

    When Diofant is working with iterables, it is almost always assuming
    that the iterable is not a string or a mapping, so those are excluded
    by default. If you want a pure Python definition, make exclude=None. To
    exclude multiple items, pass them as a tuple.

    See Also
    ========

    is_sequence

    Examples
    ========

    >>> things = [[1], (1,), {1}, Tuple(1), (j for j in [1, 2]), {1: 2}, '1', 1]
    >>> for i in things:
    ...     print(f'{iterable(i)} {type(i)}')
    True <... 'list'>
    True <... 'tuple'>
    True <... 'set'>
    True <class 'diofant.core.containers.Tuple'>
    True <... 'generator'>
    False <... 'dict'>
    False <... 'str'>
    False <... 'int'>

    >>> iterable({}, exclude=None)
    True
    >>> iterable({}, exclude=str)
    True
    >>> iterable('no', exclude=str)
    False

    """
    try:
        iter(i)
    except TypeError:
        return False
    if exclude:
        return not isinstance(i, exclude)
    return True


def is_sequence(i, include=None):
    """
    Return a boolean indicating whether ``i`` is a sequence in the Diofant
    sense. If anything that fails the test below should be included as
    being a sequence for your application, set 'include' to that object's
    type; multiple types should be passed as a tuple of types.

    Note: although generators can generate a sequence, they often need special
    handling to make sure their elements are captured before the generator is
    exhausted, so these are not included by default in the definition of a
    sequence.

    See Also
    ========

    iterable

    Examples
    ========

    >>> from types import GeneratorType
    >>> is_sequence([])
    True
    >>> is_sequence(set())
    False
    >>> is_sequence('abc')
    False
    >>> is_sequence('abc', include=str)
    True
    >>> generator = (c for c in 'abc')
    >>> is_sequence(generator)
    False
    >>> is_sequence(generator, include=(str, GeneratorType))
    True

    """
    return (hasattr(i, '__getitem__') and
            iterable(i) or
            bool(include) and
            isinstance(i, include))


def as_int(n):
    """
    Convert the argument to a builtin integer.

    The return value is guaranteed to be equal to the input. ValueError is
    raised if the input has a non-integral value.

    Examples
    ========

    >>> 3.0
    3.0
    >>> as_int(3.0)  # convert to int and test for equality
    3
    >>> int(sqrt(10))
    3
    >>> as_int(sqrt(10))
    Traceback (most recent call last):
    ...
    ValueError: ... is not an integer

    """
    try:
        result = int(n)
        if result != n:
            raise TypeError
    except TypeError as exc:
        raise ValueError(f'{n} is not an integer') from exc
    return result


# If HAS_GMPY is 0, no supported version of gmpy is available. Otherwise,
# HAS_GMPY contains the major version number of gmpy.

GROUND_TYPES: str = os.getenv('DIOFANT_GROUND_TYPES', 'auto').lower()

gmpy: typing.Any = import_module('gmpy2', min_module_version='2.1.0b5',
                                 module_version_attr='version',
                                 module_version_attr_call_args=())
if gmpy:
    HAS_GMPY = 2
else:
    HAS_GMPY = 0

if GROUND_TYPES == 'auto':
    if HAS_GMPY:
        GROUND_TYPES = 'gmpy'
    else:
        GROUND_TYPES = 'python'

if GROUND_TYPES == 'gmpy' and not HAS_GMPY:
    from warnings import warn
    warn("gmpy library is not installed, switching to 'python' ground types")
    GROUND_TYPES = 'python'

if GROUND_TYPES == 'python':
    os.environ['MPMATH_NOGMPY'] = 'yes'



"""Ground types for various mathematical domains in Diofant."""

import builtins
import fractions
from math import factorial as python_factorial
from math import gcd as python_gcd
from math import isqrt as python_sqrt

from ..core.compatibility import HAS_GMPY
from ..core.numbers import Float as DiofantReal
from ..core.numbers import Integer as DiofantInteger
from ..core.numbers import Rational as DiofantRational
from ..core.numbers import igcdex as python_gcdex


__all__ = ('python_factorial', 'python_gcd', 'python_sqrt', 'DiofantReal',
           'DiofantInteger', 'DiofantRational', 'python_gcdex', 'PythonInteger',
           'PythonInteger', 'PythonReal', 'PythonComplex', 'PythonRational',
           'gmpy_denom', 'gmpy_factorial', 'gmpy_gcd', 'gmpy_gcdex',
           'gmpy_sqrt', 'GMPYRational', 'GMPYInteger',
           'gmpy_numer', 'gmpy_qdiv')


PythonInteger = builtins.int
PythonReal = builtins.float
PythonComplex = builtins.complex
PythonRational = fractions.Fraction


if HAS_GMPY:
    # pylint: disable=no-name-in-module
    from gmpy2 import denom as gmpy_denom
    from gmpy2 import fac as gmpy_factorial
    from gmpy2 import gcd as gmpy_gcd
    from gmpy2 import gcdext as gmpy_gcdex
    from gmpy2 import isqrt as gmpy_sqrt
    from gmpy2 import mpq as GMPYRational  # noqa: N812
    from gmpy2 import mpz as GMPYInteger  # noqa: N812
    from gmpy2 import numer as gmpy_numer
    from gmpy2 import qdiv as gmpy_qdiv
else:
    class _GMPYInteger:
        def __init__(self, obj):
            pass

    class _GMPYRational:
        def __init__(self, obj):
            pass

    GMPYInteger = _GMPYInteger
    GMPYRational = _GMPYRational
    gmpy_factorial = None
    gmpy_numer = None
    gmpy_denom = None
    gmpy_gcdex = None
    gmpy_gcd = None
    gmpy_sqrt = None
    gmpy_qdiv = None
